import bisect
import json
import os
from collections import namedtuple

import botocore
import sqlalchemy
from aio_pyformance.registry import MetricsRegistry
import base64
import datetime
import logging
import re

from pscr_common.aws import get_boto3_client

from feed_manager.api.models.internal_models import ExtendedReport
from pscr_common.utils import get_utcnow_with_tzinfo
from sqlalchemy import and_, or_, tuple_, null, exists
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy import delete, select
from sqlalchemy.exc import IntegrityError
from typing import List, Dict, Union, Any, Iterable, Collection, Tuple, Optional
import uuid

from feed_manager.api.models import (
    FeedInfo as ApiFeedInfo,
    Report as ApiReport,
    Feed as ApiFeed,
    ValidationError,
)
from feed_manager.engine.errors import EntityNotFound, NotPermissioned
from feed_manager.engine.report_engine import ReportEngine
from feed_manager.sql.models import (
    Feed as DbFeed,
    Report as DbReport,
    Classifier,
    ClassifierWatchlist as DbClassifierWatchlist,
    ReportWatchlist as DbReportWatchlist,
    ReportClassifier,
    Watchlist,
    ReservedFeedDelegate,
)
from feed_manager.utils.config import Config
from feed_manager.utils.db import (
    ReadWriteDbPool,
    is_fk_violation_exception,
    is_unique_violation_exception,
)
from feed_manager.utils.delegates_client import DelegatesClient
from feed_manager.utils.mem_cache import MemCache as OrgFeatureCache
from feed_manager.utils.publisher import Publisher
from feed_manager.utils.query_converter import convert_query
from feed_manager.utils.util import str_to_bool

_logger = logging.getLogger(__name__)


class InvalidFeedType(ValidationError):
    def __init__(self, feed_id, expected_type):
        super(InvalidFeedType, self).__init__(
            f"Feed '{feed_id}' is not of type '{expected_type}'"
        )


class FeedNotFound(EntityNotFound):
    def __init__(self, feed_id):
        super(FeedNotFound, self).__init__(f"Feed '{feed_id}' not found")


class FeedNotPermissioned(NotPermissioned):
    def __init__(self, feed_id):
        super(FeedNotPermissioned, self).__init__(
            f"Not permissioned for feed '{feed_id}'"
        )


class FeedEngine(object):
    VALID_ID_RE = re.compile(r"^[\w-]+$", re.ASCII)
    FEED_REPORT_ID_SEPARATER = "-"
    FEED_ID_CLASSIFIER = "feed_id"
    S3_ADMIN_BUCKET = os.environ.get("AWS_ADMIN_BUCKET")
    MDR_FEED_NAME = os.environ.get("MDR_FEED_NAME", "MDR-FEED")

    def __init__(
        self,
        config: Config,
        db_pool: ReadWriteDbPool,
        publisher: Publisher,
        metrics: MetricsRegistry,
    ):
        self.config = config
        self.db_pool = db_pool
        self.publisher = publisher
        self.metrics = metrics

        self.report_engine = ReportEngine(config, db_pool, metrics)
        self.delegates_client = DelegatesClient(config)

    async def shutdown(self, _app=None):
        await self.report_engine.shutdown()
        await self.delegates_client.close()

    @classmethod
    def get_report_unique_id(cls, prefix, report_id):
        return f"{prefix}{cls.FEED_REPORT_ID_SEPARATER}{report_id}"

    @classmethod
    def remove_report_id_prefix(cls, report_id):
        return report_id.partition(cls.FEED_REPORT_ID_SEPARATER)[2]

    async def create_feed(
        self,
        org_id: str,
        feedinfo: ApiFeedInfo,
        reports: Iterable[ApiReport],
        delegates: Collection[str] = None,
        now: datetime.datetime = None,
    ) -> ApiFeedInfo:
        now = now if now else get_utcnow_with_tzinfo()
        with self.metrics.timer("feeds.create.duration").time():
            feed_id = self.gen_feed_id()
            reports_with_id_prefix = [
                self._with_id_prefix(feed_id, report) for report in reports
            ]

            await self.report_engine.validate_reports(org_id, reports_with_id_prefix)

            async with self.db_pool.rw_transaction() as conn:
                try:
                    await conn.execute(
                        DbFeed.__table__.insert().values(
                            **self._feed_to_db_dict(feedinfo, org_id, feed_id, now)
                        )
                    )

                    feed_info_classifiers = self._feed_to_classifiers(
                        feedinfo, org_id, feed_id, delegates
                    )
                    classifier_ids = await self.report_engine.insert_classifiers(
                        feed_info_classifiers
                    )

                    for report in reports_with_id_prefix:
                        await self.report_engine.add_report(
                            conn, report=report, classifier_ids=classifier_ids
                        )
                except IntegrityError:
                    raise ValueError(f"Feed {feedinfo.name} already exists")

                await self._set_feed_delegates(conn, feed_id, delegates)

            rtn_feed_info = ApiFeedInfo(
                id=feed_id,
                name=feedinfo.name,
                owner=feedinfo.owner,
                provider_url=feedinfo.provider_url,
                summary=feedinfo.summary,
                category=feedinfo.category,
                access=feedinfo.access,
                source_label=feedinfo.source_label,
                alertable=feedinfo.alertable,
                scope=feedinfo.scope,
            )

            await self.publisher.publish_report_notifications(
                [report.id for report in reports_with_id_prefix]
            )

            return rtn_feed_info

    async def get_feeds(
        self, org_id: str, include_public: bool = False
    ) -> List[ApiFeedInfo]:
        feeds = []

        sel_stmt = self.feedinfos_select(org_id, include_public)

        async with self.db_pool.rw_transaction() as conn:
            async for result in await conn.stream(sel_stmt):
                feeds.append(self._feedinfo_to_api(result[0], org_id=org_id))

        return self._filter_feeds(feeds, FeedEngine.get_org_features(org_id), org_id)

    async def get_feed(self, org_id: str, feed_id: str) -> ApiFeed:
        feedinfo = await self.get_feedinfo(org_id, feed_id)
        await self.verify_feedinfo_permissions(org_id, feedinfo, require_owner=False)

        reports = await self.get_feed_reports(org_id, feed_id)

        return ApiFeed(feedinfo, reports)

    async def get_feed_ids(self, feed_names: List[str]) -> List[str]:
        async with self.db_pool.ro_transaction() as conn:
            sel_stmt = select([DbFeed.id]).where(DbFeed.name.in_(feed_names))
            return [row["id"] async for row in await conn.stream(sel_stmt)]

    async def get_feed_name(self, feed_id: str, conn=None) -> str:
        if not conn:
            async with self.db_pool.ro_transaction() as conn:
                return await self._get_feed_name(conn, feed_id)
        else:
            return await self._get_feed_name(conn, feed_id)

    @staticmethod
    async def _get_feed_name(conn, feed_id: str) -> str:
        sel_stmt = select([DbFeed.name]).where(DbFeed.id == feed_id).limit(1)

        result = (await conn.execute(sel_stmt)).fetchone()

        if result is None:
            raise FeedNotFound(feed_id)

        return result["name"]

    async def get_report_title(self, feed_id: str, report_id: str) -> str:
        return await self.report_engine.get_report_title(
            self.get_report_unique_id(feed_id, report_id)
        )

    async def remove_feed(self, org_id: str, feed_id: str) -> bool:
        await self.verify_feed_permissions(org_id, feed_id, require_owner=True)

        with self.metrics.timer("feeds.delete.duration").time():
            deleted_reports = []
            async with self.db_pool.rw_transaction() as conn:
                del_stmt = (
                    delete(DbFeed)
                    .where(DbFeed.owner == org_id)
                    .where(DbFeed.id == feed_id)
                )

                result = await conn.execute(del_stmt)
                if result.rowcount == 0:
                    return False

                await self.report_engine.dirty_org_watchlist_for_classifier(
                    conn,
                    classifier_key=self.FEED_ID_CLASSIFIER,
                    classifier_value=feed_id,
                )

                # Delete all reports tied to classifier "feed_id"==feed_id
                join_stmt = ReportClassifier.__table__.join(
                    Classifier.__table__,
                    ReportClassifier.classifier_id == Classifier.id,
                )

                sel_stmt = (
                    select([ReportClassifier.report_id])
                    .select_from(join_stmt)
                    .where(Classifier.key == self.FEED_ID_CLASSIFIER)
                    .where(Classifier.value == feed_id)
                )

                reports_del_stmt = (
                    delete(DbReport)
                    .where(DbReport.id.in_(sel_stmt))
                    .returning(DbReport.id)
                    .execution_options(synchronize_session=False)
                )
                async for result in await conn.stream(reports_del_stmt):
                    bisect.insort(deleted_reports, result["id"])

                await conn.execute(
                    delete(Classifier)
                    .where(Classifier.key == self.FEED_ID_CLASSIFIER)
                    .where(Classifier.value == feed_id)
                )

                join_stmt = Watchlist.__table__.join(
                    DbClassifierWatchlist,
                    DbClassifierWatchlist.watchlist_id == Watchlist.id,
                    isouter=True,
                ).join(
                    DbReportWatchlist,
                    DbReportWatchlist.watchlist_id == Watchlist.id,
                    isouter=True,
                )
                sel_stmt = (
                    select([Watchlist.id])
                    .select_from(join_stmt)
                    .where(
                        and_(
                            DbReportWatchlist.watchlist_id == null(),
                            DbClassifierWatchlist.watchlist_id == null(),
                        )
                    )
                )

                del_stmt = (
                    delete(Watchlist)
                    .where(Watchlist.id.in_(sel_stmt))
                    .execution_options(synchronize_session=False)
                )

                # delete all the orphaned watchlists
                await conn.execute(del_stmt)

            await self.publisher.publish_report_notifications(deleted_reports)

            return True

    async def get_feedinfo(
        self, org_id: str, feed_id: str, include_reports_count=False
    ) -> ApiFeedInfo:
        async with self.db_pool.ro_transaction() as conn:
            return await self._get_feedinfo(
                conn, org_id, feed_id, include_reports_count
            )

    async def upsert_feedinfo(
        self,
        org_id: str,
        feed_id: str,
        feedinfo: ApiFeedInfo,
        now: datetime.datetime = None,
    ) -> None:
        now = now if now else get_utcnow_with_tzinfo()

        async with self.db_pool.rw_transaction() as conn:
            existing_feedinfo = await self._get_feedinfo(conn, org_id, feed_id)
            await self.verify_feedinfo_permissions(
                org_id, existing_feedinfo, require_owner=True
            )

            insert_stmt = insert(
                DbFeed,
                values={
                    "id": feedinfo.id,
                    "name": feedinfo.name,
                    "owner": org_id,
                    "provider_url": feedinfo.provider_url,
                    "summary": feedinfo.summary,
                    "category": feedinfo.category,
                    "access": feedinfo.access,
                    "source_label": feedinfo.source_label,
                    "timestamp": now,
                    "alertable": feedinfo.alertable,
                    "scope": json.dumps(feedinfo.scope) if feedinfo.scope else None,
                },
            )

            upsert_stmt = insert_stmt.on_conflict_do_update(
                index_elements=[DbFeed.id],
                set_={
                    "name": insert_stmt.excluded.name,
                    "owner": insert_stmt.excluded.owner,
                    "provider_url": insert_stmt.excluded.provider_url,
                    "summary": insert_stmt.excluded.summary,
                    "category": insert_stmt.excluded.category,
                    "access": insert_stmt.excluded.access,
                    "source_label": insert_stmt.excluded.source_label,
                    "timestamp": insert_stmt.excluded.timestamp,
                    "alertable": insert_stmt.excluded.alertable,
                    "scope": insert_stmt.excluded.scope,
                },
            )

            try:
                result = await conn.execute(upsert_stmt)

                # if updated and feedinfo content is different
                # - add new source_label classifier if necessary
                # - update report classifiers if necessary
                #   - delete old source_label report classifier
                #   - set new source_label report classifier
                # - publish reports with new feedinfo
                if result.rowcount == 1 and existing_feedinfo != feedinfo:
                    # The feedinfo changed - we need to publish updates for the reports tied to this feed
                    feed_classifier_id = await self.report_engine.get_classifier_id(
                        self.FEED_ID_CLASSIFIER, feed_id
                    )

                    # get all the reports for feed_id
                    reports = await self.report_engine.get_reports_with_classifiers(
                        [feed_classifier_id]
                    )
                    if existing_feedinfo.source_label != feedinfo.source_label:
                        if existing_feedinfo.source_label:
                            await self.report_engine.remove_report_classifiers(
                                conn,
                                report_ids=[report.id for report in reports],
                                report_classifiers=[
                                    await self.report_engine.get_classifier_id(
                                        "source_label", existing_feedinfo.source_label
                                    )
                                ],
                            )

                        if feedinfo.source_label:
                            new_classifier_ids = (
                                await self.report_engine.insert_classifiers(
                                    [
                                        {
                                            "key": "source_label",
                                            "value": feedinfo.source_label,
                                        }
                                    ]
                                )
                            )
                            await self.report_engine.insert_report_classifiers(
                                conn,
                                {report.id: new_classifier_ids for report in reports},
                            )

                    await self.publisher.publish_report_notifications(
                        [report.id for report in reports]
                    )
            except IntegrityError as e:
                if is_fk_violation_exception(e):
                    _logger.error("Unable to update feedinfo: %s", feed_id)
                    raise ValueError("Unable to update feedinfo: %s" % feed_id)
                elif is_unique_violation_exception(e):
                    raise ValueError(f"Feed {feedinfo.name} already exists")
                else:
                    raise e

    async def set_feed_delegates(
        self, org_id: str, feed_id: str, delegates: Collection[str]
    ) -> None:
        async with self.db_pool.rw_transaction() as conn:
            feedinfo = await self._get_feedinfo(conn, org_id, feed_id)
            await self.verify_feedinfo_permissions(org_id, feedinfo, require_owner=True)
            reports = await self.get_feed_reports(
                org_id, feedinfo.id, remove_id_prefix=False
            )

            deleted_delegates = await self._set_feed_delegates(conn, feed_id, delegates)

            # Delete view report classifiers for removed delegate orgs
            if deleted_delegates:
                view_classifiers = [
                    await self.report_engine.get_classifier_id(
                        ReportEngine.VIEW_CLASSIFIER, delegate_org
                    )
                    for delegate_org in deleted_delegates
                ]
                await self.report_engine.remove_report_classifiers(
                    conn,
                    report_ids=[report.id for report in reports],
                    report_classifiers=view_classifiers,
                )

            # Make sure view report classifier exists for all delegates
            new_classifier_ids = await self.report_engine.insert_classifiers(
                [
                    {"key": ReportEngine.VIEW_CLASSIFIER, "value": delegate_org}
                    for delegate_org in delegates
                ]
            )
            await self.report_engine.insert_report_classifiers(
                conn, {report.id: new_classifier_ids for report in reports}
            )

        await self.publisher.publish_report_notifications(
            [report.id for report in reports]
        )

    async def _set_feed_delegates(
        self, conn, feed_id: str, delegates: Collection[str]
    ) -> List[str]:
        del_stmt = (
            delete(ReservedFeedDelegate)
            .where(ReservedFeedDelegate.feed_id == feed_id)
            .returning(ReservedFeedDelegate.org_id)
        )

        if delegates:
            insert_stmt = insert(ReservedFeedDelegate).values(
                [{"feed_id": feed_id, "org_id": org_id} for org_id in delegates]
            )

            upsert_stmt = insert_stmt.on_conflict_do_nothing(
                index_elements=[
                    ReservedFeedDelegate.feed_id,
                    ReservedFeedDelegate.org_id,
                ]
            )

            await conn.execute(upsert_stmt)

            # If there were reserved orgs we only want to delete those not in the list
            del_stmt = del_stmt.where(~ReservedFeedDelegate.org_id.in_(delegates))

        org_ids = []
        async for row in await conn.stream(del_stmt):
            org_ids.append(row["org_id"])

        return org_ids

    async def get_feed_delegates(self, org_id: str, feed_id: str) -> List[str]:
        async with self.db_pool.rw_transaction() as conn:
            feedinfo = await self._get_feedinfo(conn, org_id, feed_id)
            await self.verify_feedinfo_permissions(org_id, feedinfo, require_owner=True)

            if feedinfo.access != ReportEngine.RESERVED_LABEL:
                raise InvalidFeedType(feed_id, ReportEngine.RESERVED_LABEL)

            return await self._get_feed_delegates(feed_id)

    async def _get_feed_delegates(self, feed_id: str) -> List[str]:
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = select([ReservedFeedDelegate.org_id]).where(
                ReservedFeedDelegate.feed_id == feed_id
            )

            ids = []
            async for row in await conn.stream(sel_stmt):
                ids.append(row["org_id"])

            return ids

    async def get_feed_reports(
        self,
        org_id: str,
        feed_id: str,
        remove_id_prefix=True,
        skip=0,
        take=0,
        get_count_only: bool = False,
    ) -> Union[List[ApiReport], int]:
        await self.verify_feed_permissions(org_id, feed_id, require_owner=False)

        feed_classifier_id = await self.report_engine.get_classifier_id(
            self.FEED_ID_CLASSIFIER, feed_id
        )

        reports = await self.report_engine.get_reports_with_classifiers(
            [feed_classifier_id], skip=skip, take=take, get_count_only=get_count_only
        )

        if get_count_only or not remove_id_prefix:
            return reports

        # We remove the report id prefix (feed_id) because this 'get' is in the context of a feed.
        return [self._remove_id_prefix(report) for report in reports]

    async def get_feed_report(
        self, org_id: str, feed_id: str, feed_report_id: str
    ) -> ApiReport:
        await self.verify_feed_permissions(org_id, feed_id, require_owner=False)

        report = await self.report_engine.get_report(
            org_id, self.get_report_unique_id(feed_id, feed_report_id)
        )

        # We remove the report id prefix (feed_id) because this 'get' is in the context of a feed.
        return self._remove_id_prefix(report)

    async def get_extended_feed_reports(
        self, report_ids: Iterable[str]
    ) -> Dict[str, ExtendedReport]:
        report_feed_ids = await self.report_engine.get_report_classifier_values(
            report_ids, FeedEngine.FEED_ID_CLASSIFIER
        )
        extended_reports = {}

        if report_feed_ids:
            feed_ids = set(report_feed_ids.values())
            feeds = await self._get_feedinfos_for_reports(feed_ids)

            delegate_orgs = {
                feed_id: await self._get_feed_delegates(feed_id) for feed_id in feed_ids
            }

            reports = await self.report_engine.get_reports_from_ids(
                report_feed_ids.keys()
            )

            for report in reports:
                feedinfo = feeds.get(report_feed_ids[report.id])
                if feedinfo:
                    extended_reports[report.id] = ExtendedReport.from_basic(
                        report,
                        source=feedinfo.owner,
                        access=feedinfo.access,
                        feedinfo=feedinfo,
                        delegates=delegate_orgs.get(feedinfo.id),
                    )
                else:
                    _logger.info(
                        "get_extended_feed_reports: Missing feedinfo for report %s",
                        report.id,
                    )

        return extended_reports

    async def replace_reports(
        self,
        org_id: str,
        feed_id: str,
        reports: Iterable[ApiReport],
        now: datetime.datetime = None,
    ) -> Tuple[int, int, int]:
        now = now if now else get_utcnow_with_tzinfo()

        feedinfo = await self.get_feedinfo(org_id, feed_id)
        await self.verify_feedinfo_permissions(org_id, feedinfo, require_owner=True)

        feed_info_classifiers = self._feed_to_classifiers(
            feedinfo, org_id, feed_id, await self._get_feed_delegates(feed_id)
        )
        classifier_ids = await self.get_classifier_ids(feed_info_classifiers)
        classifier_ids.extend(await self.get_feed_classifier_ids(feed_id))

        await self.report_engine.validate_reports(org_id, reports)

        reports_with_id_prefix = [
            self._with_id_prefix(feed_id, report) for report in reports
        ]

        async with self.db_pool.rw_transaction() as conn:
            # Remove old reports
            join_stmt = DbReport.__table__.join(
                ReportClassifier.__table__, ReportClassifier.report_id == DbReport.id
            ).join(
                Classifier.__table__, ReportClassifier.classifier_id == Classifier.id
            )
            # sel_stmt matches all existing reports for feed_id
            sel_stmt = (
                select([DbReport.id])
                .select_from(join_stmt)
                .where(Classifier.key == self.FEED_ID_CLASSIFIER)
                .where(Classifier.value == feed_id)
            )

            # The report ids we need to delete
            deleted_reports_ids = {r.id async for r in await conn.stream(sel_stmt)} - {
                r.id for r in reports_with_id_prefix
            }

            (
                inserted_report_ids,
                updated_report_ids,
            ) = await self.report_engine.replace_reports(
                conn, reports_with_id_prefix, classifier_ids
            )

            # dirty the watchlists before we delete them so that we don't lose the association
            await self.report_engine.dirty_org_watchlist_for_reports(
                conn, inserted_report_ids | updated_report_ids | deleted_reports_ids
            )

            await self.report_engine.delete_reports(
                conn, deleted_reports_ids, dirty_org_watchlists=False
            )

            # Update the timestamp on the feed if it is older than what we've got
            update_stmt = (
                DbFeed.__table__.update()
                .values(timestamp=now)
                .where(DbFeed.owner == org_id)
                .where(DbFeed.id == feed_id)
                .where(DbFeed.timestamp < now)
            )
            await conn.execute(update_stmt)

        await self.publisher.publish_report_notifications(
            deleted_reports_ids | inserted_report_ids | updated_report_ids
        )

        return (
            len(inserted_report_ids),
            len(updated_report_ids),
            len(deleted_reports_ids),
        )

    async def delete_report(
        self,
        org_id: str,
        feed_id: str,
        feed_report_id: str,
        now: datetime.datetime = None,
    ) -> bool:
        now = now if now else get_utcnow_with_tzinfo()

        await self.verify_feed_permissions(org_id, feed_id, require_owner=True)

        with self.metrics.timer("reports.delete.duration").time():
            async with self.db_pool.rw_transaction() as conn:
                if await self.report_engine.delete_report(
                    conn, self.get_report_unique_id(feed_id, feed_report_id)
                ):
                    # Update the timestamp on the feed if it is older than what we've got
                    del_stmt = (
                        DbFeed.__table__.update()
                        .values(timestamp=now)
                        .where(DbFeed.owner == org_id)
                        .where(DbFeed.id == feed_id)
                        .where(DbFeed.timestamp < now)
                    )
                    await conn.execute(del_stmt)

                    result = True
                else:
                    result = False

        if result:
            await self.publisher.publish_report_notifications(
                [self.get_report_unique_id(feed_id, feed_report_id)]
            )
            return True
        else:
            return False

    async def upsert_report(
        self,
        org_id: str,
        feed_id: str,
        report: ApiReport,
        now: datetime.datetime = None,
    ) -> ApiReport:
        now = now if now else get_utcnow_with_tzinfo()

        await self.report_engine.validate_reports(org_id, [report])

        feedinfo = await self.get_feedinfo(org_id, feed_id)
        await self.verify_feedinfo_permissions(org_id, feedinfo, require_owner=True)

        feed_info_classifiers = self._feed_to_classifiers(
            feedinfo, org_id, feed_id, await self._get_feed_delegates(feed_id)
        )
        classifier_ids = await self.get_classifier_ids(feed_info_classifiers)
        classifier_ids.extend(await self.get_feed_classifier_ids(feed_id))

        report_with_id_prefix = self._with_id_prefix(feed_id, report)
        with self.metrics.timer("reports.upsert.duration").time():
            async with self.db_pool.rw_transaction() as conn:
                await self.report_engine.upsert_report(
                    conn, report_with_id_prefix, classifier_ids
                )
                # Update the timestamp on the feed if it is older than what we've got
                await conn.execute(
                    DbFeed.__table__.update()
                    .values(timestamp=now)
                    .where(DbFeed.owner == org_id)
                    .where(DbFeed.id == feed_id)
                    .where(DbFeed.timestamp < now)
                )

            await self.publisher.publish_report_notifications(
                [report_with_id_prefix.id]
            )

        return report

    async def get_classifier_ids(
        self, classifiers: Collection[Dict[str, str]]
    ) -> List[int]:
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = select([Classifier.id]).where(
                tuple_(Classifier.key, Classifier.value).in_(
                    [
                        (classifier["key"], classifier["value"])
                        for classifier in classifiers
                    ]
                )
            )

            ids = []
            async for row in await conn.stream(sel_stmt):
                ids.append(row["id"])

            return ids

    async def get_feed_classifier_ids(self, feed_id: str) -> List[int]:
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = select([Classifier.id]).where(
                Classifier.key.startswith(f"feed_{feed_id}")
            )

            ids = []
            async for row in await conn.stream(sel_stmt):
                ids.append(row["id"])

            return ids

    async def publish_org_reports(
        self, org_id: str
    ) -> List[Dict[str, Union[str, int]]]:
        feed_classifier_ids = []
        results = []

        # Publish all report tied to feeds for the org
        feedinfos = await self.get_feeds(org_id)
        for feedinfo in feedinfos:
            feed_classifier_ids.append(
                await self.report_engine.get_classifier_id(
                    self.FEED_ID_CLASSIFIER, feedinfo.id
                )
            )

            reports = await self.get_feed_reports(
                org_id, feedinfo.id, remove_id_prefix=False
            )

            await self.publisher.publish_report_notifications(
                [report.id for report in reports]
            )
            results.append({"feed_id": feedinfo.id, "num_reports": len(reports)})

        # publish all reports not tied to feeds for the org
        try:
            source_classifier = await self.report_engine.get_classifier_id(
                "source", org_id
            )
            reports = await self.report_engine.get_reports_with_classifiers(
                classifier_ids=[source_classifier],
                exclude_classifier_ids=feed_classifier_ids,
            )
            await self.publisher.publish_report_notifications(
                [report.id for report in reports]
            )
            results.append({"feed_id": None, "num_reports": len(reports)})
        except ValueError:
            # It's possible there is no classifier for source/org_id
            pass

        return results

    async def publish_org_reports_for_feed(self, org_id: str, feed_id: str) -> int:
        # Publish all report tied to feeds for the org
        feedinfo = await self.get_feedinfo(org_id, feed_id)
        reports = await self.get_feed_reports(
            org_id, feedinfo.id, remove_id_prefix=False
        )

        await self.publisher.publish_report_notifications(
            [report.id for report in reports]
        )

        return len(reports)

    async def translate_legacy_query(self, query: str) -> str:
        converted_query = convert_query(query)

        # org key doesn't matter here because we use all stage 0 searchers across orgs to validate
        await self.report_engine.validate_queries(
            "translate_legacy_query_org", [converted_query]
        )

        return converted_query

    async def _get_feedinfos_for_reports(
        self, feed_ids: Iterable[str]
    ) -> Dict[str, ApiFeedInfo]:
        sel_stmt = select([DbFeed]).where(DbFeed.id.in_(feed_ids))

        async with self.db_pool.rw_transaction() as conn:
            return {
                result[0].id: self._feedinfo_to_api(result[0], org_id=result[0].owner)
                async for result in await conn.stream(sel_stmt)
            }

    @staticmethod
    def gen_feed_id() -> str:
        return (
            base64.b64encode(uuid.uuid4().bytes)
            .decode("utf-8")
            .rstrip("=\n")
            .replace("/", "")
            .replace("+", "")
        )

    @staticmethod
    def feedinfos_select(
        org_id: str, include_public: bool = False
    ) -> sqlalchemy.sql.expression.select:
        join_stmt = DbFeed.__table__.outerjoin(
            ReservedFeedDelegate.__table__,
            and_(
                DbFeed.id == ReservedFeedDelegate.feed_id,
                ReservedFeedDelegate.org_id == org_id,
            ),
        )

        sel_stmt = select([DbFeed]).select_from(join_stmt)

        if include_public:
            sel_stmt = sel_stmt.where(
                or_(
                    DbFeed.owner == org_id,
                    DbFeed.access == ReportEngine.PUBLIC_LABEL,
                    and_(
                        DbFeed.access == ReportEngine.RESERVED_LABEL,
                        ReservedFeedDelegate.org_id == org_id,
                    ),
                )
            )
        else:
            sel_stmt = sel_stmt.where(
                or_(
                    DbFeed.owner == org_id,
                    and_(
                        DbFeed.access == ReportEngine.RESERVED_LABEL,
                        ReservedFeedDelegate.org_id == org_id,
                    ),
                )
            )

        return sel_stmt

    async def _get_feedinfo(
        self, conn, org_id: str, feed_id: str, include_reports_count=False
    ) -> ApiFeedInfo:
        sel_stmt = (
            self.feedinfos_select(org_id=org_id, include_public=True)
            .where(DbFeed.id == feed_id)
            .limit(1)
        )

        result = (await conn.execute(sel_stmt)).scalar()

        if not result:
            raise FeedNotFound(feed_id)

        return self._feedinfo_to_api(
            result,
            org_id,
            reports_count=await self.get_feed_reports(
                org_id, feed_id, get_count_only=True
            )
            if include_reports_count
            else None,
        )

    async def verify_feed_permissions(
        self, org_id: str, feed_id: str, require_owner: bool
    ) -> None:
        async with self.db_pool.ro_transaction() as conn:
            result = (
                await conn.execute(
                    select([DbFeed]).where(DbFeed.id == feed_id).limit(1)
                )
            ).scalar()

            if not result:
                raise FeedNotFound(feed_id)

            await self.verify_feedinfo_permissions(
                org_id,
                self._feedinfo_to_api(result, org_id=org_id),
                require_owner=require_owner,
            )

    async def verify_feedinfo_permissions(
        self, org_id: str, feedinfo: ApiFeedInfo, require_owner: bool
    ) -> None:
        if feedinfo.owner != org_id:
            if require_owner:
                if feedinfo.access == ReportEngine.PUBLIC_LABEL:
                    raise FeedNotPermissioned(feedinfo.id)
                elif (
                    feedinfo.access == ReportEngine.RESERVED_LABEL
                    and org_id in await self._get_feed_delegates(feedinfo.id)
                ):
                    raise FeedNotPermissioned(feedinfo.id)
                else:
                    raise FeedNotFound(feedinfo.id)
            elif feedinfo.access == ReportEngine.PRIVATE_LABEL:
                # Feed is private and we needed org_id to be owner
                raise FeedNotFound(feedinfo.id)

    @classmethod
    def _feedinfo_to_api(
        cls, record: DbFeed, org_id: str, reports_count: Optional[int] = None
    ) -> ApiFeedInfo:
        if record.owner == org_id:
            owner = record.owner
        elif record.access in (ReportEngine.PUBLIC_LABEL, ReportEngine.RESERVED_LABEL):
            owner = record.source_label if record.source_label else None
        else:
            owner = None

        return ApiFeedInfo(
            id=record.id,
            name=record.name,
            owner=owner,
            provider_url=record.provider_url,
            summary=record.summary,
            category=record.category,
            access=record.access,
            source_label=record.source_label,
            alertable=record.alertable,
            reports_count=reports_count,
            scope=json.loads(record.scope) if record.scope else None,
        )

    @staticmethod
    def _feed_to_db_dict(
        record: ApiFeedInfo, org_id: str, feed_id: str, now: datetime.datetime
    ) -> Dict[str, Any]:
        feed = DbFeed()
        feed.id = feed_id
        feed.name = record.name
        feed.owner = org_id
        feed.source_label = record.source_label
        feed.provider_url = record.provider_url
        feed.summary = record.summary
        feed.category = record.category
        feed.access = record.access
        feed.timestamp = now
        feed.alertable = record.alertable
        feed.scope = json.dumps(record.scope) if record.scope else None

        return feed.as_dict()

    @classmethod
    def _feed_to_classifiers(
        cls,
        info: ApiFeedInfo,
        org_id: str,
        feed_id: str,
        delegates: Collection[str] = None,
    ) -> List[Dict[str, str]]:
        d = [
            {"key": cls.FEED_ID_CLASSIFIER, "value": feed_id},
            {"key": ReportEngine.ACCESS_CLASSIFIER, "value": info.access},
            {"key": ReportEngine.SOURCE_CLASSIFIER, "value": org_id},
        ]

        if info.source_label:
            d.append({"key": "source_label", "value": info.source_label})

        if delegates:
            for delegate in delegates:
                d.append({"key": ReportEngine.VIEW_CLASSIFIER, "value": delegate})

        return d

    @classmethod
    def _remove_id_prefix(cls, report: ApiReport) -> ApiReport:
        return ApiReport(
            id=cls.remove_report_id_prefix(report.id),
            timestamp=report.timestamp,
            title=report.title,
            description=report.description,
            severity=report.severity,
            link=report.link,
            tags=report.tags,
            iocs=report.iocs,
            iocs_v2=report.iocs_v2,
            iocs_total_count=report.iocs_total_count,
            visibility=report.visibility,
        )

    @classmethod
    def _with_id_prefix(cls, prefix: str, report: ApiReport) -> ApiReport:
        return ApiReport(
            id=cls.get_report_unique_id(prefix, report.id),
            timestamp=report.timestamp,
            title=report.title,
            description=report.description,
            severity=report.severity,
            link=report.link,
            tags=report.tags,
            iocs=report.iocs,
            iocs_v2=report.iocs_v2,
            iocs_total_count=report.iocs_total_count,
            visibility=report.visibility,
        )

    @staticmethod
    def _filter_feeds(feeds: List, features: List, owner: str) -> List[ApiFeedInfo]:
        filtered_feeds = []
        for feed in feeds:
            if feed.scope is None or feed.owner == owner:
                filtered_feeds.append(feed)
            elif all(scope in features for scope in feed.scope):
                filtered_feeds.append(feed)

        return filtered_feeds

    @staticmethod
    def get_org_features(org_id):
        org_feature_cache_enabled = str_to_bool(
            os.environ.get("FEATURES_CACHE_ENABLED", "false")
        )
        org_feature_cache = OrgFeatureCache().get("org_features")
        if org_feature_cache_enabled and bool(org_feature_cache):
            _logger.info("Getting Org Features from Cache")
            if org_feature_cache.get(org_id):
                return org_feature_cache[org_id]
        else:
            _logger.info("Getting Org Features from S3")
            try:
                s3_client = get_boto3_client("s3")
                org_features = (
                    s3_client.get_object(
                        Bucket=FeedEngine.S3_ADMIN_BUCKET,
                        Key="org_features/features.json",
                    )["Body"]
                    .read()
                    .decode("utf-8")
                )
                org_features_dict = json.loads(org_features)
                if org_features_dict.get(org_id):
                    return org_features_dict[org_id]
            except botocore.exceptions.ClientError as error:
                _logger.error(
                    f"Error downloading org features due to {error.response['Error']['Code']}"
                )

        return []

    @staticmethod
    def remove_empty_feed_scope(feed):
        if feed.scope is None:
            new_feed = feed._asdict()
            del new_feed["scope"]
            return namedtuple("FeedInfo", new_feed.keys())(*new_feed.values())

        return feed
