import logging
import rapidjson
from aiohttp import web
from aiohttp.web_exceptions import HTTPBadRequest

from feed_manager.api.models.feed_models import FeedInfo, Report, Feed
from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.report_engine import ReportEngine
from feed_manager.web.route_helpers import (
    get_request_org,
    make_response_json,
    create_feed,
    validate_feed_scope,
)
from feed_manager.web.auth import Resource
from pscr_common.web.auth import AuthRouteTableDef, Action
from pscr_common.web.auditlog import call_audit_log

_logger = logging.getLogger(__name__)


# description: This endpoint will return all feeds. By default this will only include feeds for the requesting org.
#   If 'include_public' param is true this will also return all public feeds.


def build_route_table(engine: FeedEngine) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    @routes.get("/healthcheck")
    async def healthcheck(request: web.Request):
        """
        ---
        summary: Healthcheck
        description: Successful response indicates service reachability.
        produces:
        - application/octet-stream
        responses:
            204:
                description: service is available
        """
        request.do_not_log = True

        return make_response_json()

    @routes.post("/v1/feed", Resource.feeds, [Action.create])
    async def create_private_feed(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Create a new private feed
        description: Create new private feed.  Unique feed ID will be assigned by the service.  All IOCs will be converted to IOC_V2.  This feed will be owned by the caller. The feed will be available to only the org that created it.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: body
            name: body
            description: Feed object
            required: true
            schema:
              $ref: '#/definitions/Feed'
        responses:
            200:
                description: Feed created.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: invalid feed request.
        """
        params = await request.json(loads=rapidjson.loads)
        org_id = get_request_org(request)
        rsp = await create_feed(
            engine, org_key=org_id, params=params, required_access="private"
        )

        await call_audit_log(
            request,
            description=f"Created new private feed {params['feedinfo']['name']} (v1)",
        )
        return rsp

    @routes.post("/v1/feed/public", Resource.public_feeds, [Action.create])
    async def create_public_feed(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Create a new public feed
        description: Create public feed.  Unique feed ID will be assigned by the service.  All IOCs will be converted to IOC_V2.  This feed will be owned by the caller.  The feed will be available to all organizations.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: body
            name: body
            description: Feed object
            required: true
            schema:
              $ref: '#/definitions/Feed'
        responses:
            200:
                description: Feed created.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: invalid feed request.
        """
        params = await request.json(loads=rapidjson.loads)
        org_id = get_request_org(request)

        rsp = await create_feed(
            engine, org_key=org_id, params=params, required_access="public"
        )

        await call_audit_log(
            request,
            description=f"Created new public feed {params['feedinfo']['name']} (v1)",
        )
        return rsp

    @routes.get("/v1/feed", Resource.feeds, [Action.read], allow_head=False)
    async def get_feeds(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Get all feeds
        description: Retrieve all feeds owned by the caller.  Provide `include_public=true` parameter to also include public community feeds.
        produces:
        - application/json
        parameters:
          - in: query
            name: include_public
            description: Include public feeds in listing
            required: false
            type: boolean
        responses:
            200:
                description: Array of feed infos
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/FeedInfo'
        """
        include_public = request.query.get("include_public", "false").lower() == "true"

        feeds = await engine.get_feeds(get_request_org(request), include_public)

        updated_feeds = []
        for feed in feeds:
            updated_feeds.append(FeedEngine.remove_empty_feed_scope(feed))

        return make_response_json({"results": updated_feeds})

    @routes.get("/v1/feed/{feed_id}", Resource.feeds, [Action.read], allow_head=False)
    async def get_feed(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Get feed
        description: Retrieve feed with `feed_id`.  This feed must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            200:
                description: Feed
                schema:
                  $ref: '#/definitions/Feed'
        """
        feed_id = request.match_info["feed_id"]
        feed = await engine.get_feed(get_request_org(request), feed_id)
        return make_response_json(
            Feed(FeedEngine.remove_empty_feed_scope(feed.feedinfo), feed.reports)
        )

    @routes.delete("/v1/feed/{feed_id}", Resource.feeds, [Action.delete])
    async def remove_feed(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Delete feed
        description: Delete feed with `feed_id`.  This feed must be owned by the caller.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            204:
                description: Feed deleted
            400:
                description: Unknown feed.
        """
        feed_id = request.match_info["feed_id"]
        feed_name = await engine.get_feed_name(feed_id)

        await engine.remove_feed(get_request_org(request), feed_id)

        await call_audit_log(
            request, description=f"Removed feed, '{feed_name}' (ID: {feed_id}) (v1)"
        )
        return make_response_json()

    @routes.get(
        "/v1/feed/{feed_id}/feedinfo", Resource.feeds, [Action.read], allow_head=False
    )
    async def get_feedinfo(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Get feed info
        description: Retrieve feed info metadata for feed with `feed_id`.  This feed must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
        responses:
            200:
                description: Feed Info
                schema:
                  $ref: '#/definitions/FeedInfo'
        """
        feed_id = request.match_info["feed_id"]

        feedinfo = await engine.get_feedinfo(get_request_org(request), feed_id)

        return make_response_json(FeedEngine.remove_empty_feed_scope(feedinfo))

    @routes.put("/v1/feed/{feed_id}/feedinfo", Resource.feeds, [Action.update])
    async def update_feedinfo(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Update feed info
        description: Update feed info metadata for feed with `feed_id`.  This feed must be owned by the caller. Access may not be changed from initial value.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: feed id to update
            required: true
            type: string
          - in: body
            name: body
            description: FeedInfo object
            required: true
            schema:
              $ref: '#/definitions/FeedInfo'
        responses:
            200:
                description: Feed info updated.
                schema:
                  $ref: '#/definitions/FeedInfo'
            400:
                description: invalid feed request.
        """
        feed_id = request.match_info["feed_id"]
        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        if "owner" not in params:
            params["owner"] = org_id

        params["id"] = feed_id

        feed_scope = params.get("scope", None)

        if params["access"] != ReportEngine.PUBLIC_LABEL and feed_scope is not None:
            raise HTTPBadRequest(
                reason=f"feedinfo.scope is not applicable for private or reserved feed"
            )

        # TODO - uncomment once we figure out how to delete the reports
        #  associated with entitled feed from feed search after feed is updated with scope
        # validate_feed_scope(feed_scope)

        # cannot change access on a report because we need to gate who can create public feeds
        current_feedinfo = await engine.get_feedinfo(get_request_org(request), feed_id)

        # Existing non scoped feed can not be updated to scoped feed
        if feed_scope is not None and current_feedinfo.scope is None:
            raise HTTPBadRequest(
                reason=f"Currently Updating feed with feed scope is not supported."
            )

        if feed_scope is not None and current_feedinfo.scope is not None:
            if sorted(feed_scope) != sorted(current_feedinfo.scope):
                raise HTTPBadRequest(
                    reason=f"Scope can not be modified on the existing feed."
                )

        if params["access"] != current_feedinfo.access:
            raise HTTPBadRequest(reason="Access does not match current access level")

        # So we don't accidentally overwrite un-alertable status if it was omitted
        if params.get("alertable") is None and current_feedinfo.alertable is False:
            params["alertable"] = False

        feedinfo = FeedInfo.from_dict(params)

        await engine.upsert_feedinfo(org_id, feed_id, feedinfo)
        feed_name = feedinfo.name

        await call_audit_log(
            request, description=f"Updated feed, '{feed_name}' (ID: {feed_id}) (v1)"
        )
        return make_response_json(FeedEngine.remove_empty_feed_scope(feedinfo))

    @routes.get(
        "/v1/feed/{feed_id}/report", Resource.feeds, [Action.read], allow_head=False
    )
    async def get_reports(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Get reports
        description: Retrieve all the reports for feed with `feed_id`.  Feed must be owned by the caller or public.
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: feed ID for reports
            required: true
            type: string
        responses:
            200:
                description: Reports array
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/Report'
        """
        feed_id = request.match_info["feed_id"]

        reports = await engine.get_feed_reports(get_request_org(request), feed_id)

        return make_response_json({"results": reports})

    @routes.get(
        "/v1/feed/{feed_id}/report/{report_id}",
        Resource.feeds,
        [Action.read],
        allow_head=False,
    )
    async def get_report(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Get report
        description: Return report with `report_id` for feed.  Feed must be owned by the caller or public.
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            200:
                description: Report
                schema:
                  properties:
                    report:
                      $ref: '#/definitions/Report'
        """
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]

        report = await engine.get_feed_report(
            get_request_org(request), feed_id, report_id
        )

        return make_response_json({"report": report})

    @routes.post("/v1/feed/{feed_id}/report", Resource.feeds, [Action.create])
    async def replace_reports(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Replace reports.
        description: Replace reports for feed ID. All IOCs will be converted to IOC_V2. Any existing reports not in the payload will be deleted.  Feed must be owned by the caller.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: body
            name: body
            description: Reports list
            required: true
            schema:
              properties:
                reports:
                  type: array
                  items:
                    $ref: '#/definitions/Report'
        responses:
            200:
                description: success
                schema:
                  type: object
                  required:
                  - success
                  properties:
                    success:
                      type: boolean
        """
        feed_id = request.match_info["feed_id"]
        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        reports = [Report.from_dict(report) for report in params["reports"]]

        inserted, updated, deleted = await engine.replace_reports(
            org_id, feed_id, reports
        )

        reports_info = []
        for report in reports:
            reports_info.append(f"'{report.title}' (ID: {report.id})")

        feed_name = await engine.get_feed_name(feed_id)

        await call_audit_log(
            request,
            description=f"Replaced reports, {', '.join(reports_info)} in feed, '{feed_name}' (ID: {feed_id}) (v1)",
        )
        return make_response_json({"success": inserted + updated + deleted > 0})

    @routes.put(
        "/v1/feed/{feed_id}/report/{report_id}", Resource.feeds, [Action.update]
    )
    async def update_report(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Update report
        description: Update report with `report_id` for feed.  All IOCs will be converted to IOC_V2.  Feed must be owned by the caller.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: body
            name: body
            description: Report object
            required: true
            schema:
              $ref: '#/definitions/Report'
        responses:
            200:
                description: Report
                schema:
                  $ref: '#/definitions/Report'
            400:
                description: Unable to update report because provided timestamp is older than existing timestamp
        """
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]

        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        params["id"] = report_id

        report = await engine.upsert_report(org_id, feed_id, Report.from_dict(params))

        report_title = params["title"]
        if report_title is None:
            report_title = report.title
        feed_name = await engine.get_feed_name(feed_id)

        await call_audit_log(
            request,
            description=f"Updated report, '{report_title}' (ID: {report_id}) in feed, '{feed_name}' (ID: {feed_id}) (v1)",
        )
        return make_response_json(report)

    @routes.delete(
        "/v1/feed/{feed_id}/report/{report_id}", Resource.feeds, [Action.delete]
    )
    async def remove_report(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Delete report
        description: Delete report with `report_id` for feed .  Feed must be owned by the caller.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: feed_id
            description: ID for feed
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            204:
                description: report deleted
        """
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]

        org_id = get_request_org(request)

        report_title = await engine.get_report_title(feed_id, report_id)
        feed_name = await engine.get_feed_name(feed_id)

        await engine.delete_report(org_id, feed_id, report_id)

        await call_audit_log(
            request,
            description=f"Deleted report, '{report_title}' (ID: {report_id}) in feed, '{feed_name}' (ID: {feed_id}) (v1)",
        )
        return make_response_json()

    @routes.post("/v1/query/translate", Resource.feeds, [Action.read])
    async def translate_query(request: web.Request):
        """
        ---
        tags:
        - v1
        summary: Convert legacy query
        description: Convert CB EDR query to CBC Enterprise EDR query.  This will adjust field names and other syntax to match CBC Enterprise EDR Solr requirements.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: body
            name: body
            description: Legacy query
            required: true
            schema:
                type: object
                required:
                - query
                properties:
                  query:
                    type: string
        responses:
            200:
                description: Translated query
                schema:
                  type: object
                  properties:
                    query:
                      type: string
            400:
                description: Unable to convert query due to incompatible fields
        """
        params = await request.json(loads=rapidjson.loads)

        translated_query = await engine.translate_legacy_query(params["query"])

        translated_query = "(" + translated_query + ") -enriched:true"

        return make_response_json({"query": translated_query})

    return routes
